from fastapi import HTTPException
from starlette.status import HTTP_401_UNAUTHORIZED

InvalidCredentialsException = HTTPException(
    status_code=HTTP_401_UNAUTHORIZED,
    detail="Invalid credentials",
    headers={"WWW-Authenticate": "Bearer"}
)



from datetime import timedelta
import time
from http.cookies import SimpleCookie
from unittest.mock import Mock, AsyncMock

import pytest
from fastapi import HTTPException
from starlette.responses import Response

from fastapi_login import LoginManager
from fastapi_login.exceptions import InvalidCredentialsException


@pytest.mark.asyncio
async def test_token_expiry(clean_manager, default_data):
    token = clean_manager.create_access_token(
        data=default_data,
        expires=timedelta(microseconds=1)  # should be invalid instantly
    )
    time.sleep(1)
    with pytest.raises(HTTPException) as exc_info:
        await clean_manager.get_current_user(token)

    assert exc_info


@pytest.mark.asyncio
@pytest.mark.parametrize("loader", [Mock(), AsyncMock()])
async def test_user_loader(loader, clean_manager, default_data, db):
    token = clean_manager.create_access_token(data=default_data)
    loader = Mock()
    clean_manager.user_loader(loader)
    _ = await clean_manager.get_current_user(token)

    loader.assert_called()
    loader.assert_called_with(default_data['sub'])


@pytest.mark.asyncio
async def test_user_loader_not_set(clean_manager, default_data):
    token = clean_manager.create_access_token(data=default_data)
    with pytest.raises(Exception) as exc_info:
        await clean_manager.get_current_user(token)

    assert "Missing user_loader callback" == str(exc_info.value)


@pytest.mark.asyncio
async def test_user_loader_returns_none(clean_manager, invalid_data, load_user_fn):
    clean_manager.user_loader(load_user_fn)
    token = clean_manager.create_access_token(data={"sub": invalid_data["username"]})
    with pytest.raises(HTTPException) as exc_info:
        await clean_manager.get_current_user(token)

    assert exc_info.value == InvalidCredentialsException


def test_token_from_cookie(clean_manager):
    request = Mock(cookies={clean_manager.cookie_name: "test-value"})
    token = clean_manager._token_from_cookie(request)
    assert token == "test-value"


def test_token_from_cookie_raises(clean_manager):
    request = Mock(cookies={clean_manager.cookie_name: ""})
    with pytest.raises(HTTPException) as exc_info:
        clean_manager._token_from_cookie(request)

    assert exc_info.value == InvalidCredentialsException
    
    
def test_token_from_cookie_returns_none_auto_error_off(clean_manager):
    clean_manager.auto_error = False
    request = Mock(cookies={clean_manager.cookie_name: ""})
    token = clean_manager._token_from_cookie(request)
    assert token is None


def test_set_cookie(clean_manager, default_data):
    token = clean_manager.create_access_token(data=default_data)
    response = Response()
    clean_manager.set_cookie(response, token)
    cookie = SimpleCookie(response.headers['set-cookie'])
    cookie_value = cookie.get(clean_manager.cookie_name)
    assert cookie_value is not None
    assert cookie_value["httponly"] is True
    assert cookie_value["samesite"] == "lax"
    assert cookie_value.value == token
    assert cookie_value.key == clean_manager.cookie_name


def test_config_no_cookie_no_header_raises(secret, token_url):
    with pytest.raises(Exception) as exc_info:
        LoginManager(secret, token_url, use_cookie=False, use_header=False)

    assert "use_cookie and use_header are both False one of them needs to be True" == str(exc_info.value)






from unittest.mock import Mock

import pytest
from fastapi import Depends
from starlette.responses import RedirectResponse

from fastapi_login import LoginManager
from fastapi_login.exceptions import InvalidCredentialsException


@pytest.fixture
def exception_manager(app, secret, token_url, load_user_fn, custom_exception) -> LoginManager:
    instance = LoginManager(secret, token_url)
    instance.user_loader(load_user_fn)

    # exception handling setup

    def redirect_on_auth_exc(request, exc):
        return RedirectResponse(url="/redirect")

    instance.not_authenticated_exception = custom_exception
    app.add_exception_handler(custom_exception, redirect_on_auth_exc)

    # routes
    @app.get("/private/exception")
    def raise_exception(_=Depends(instance)):
        return {"detail": "error"}

    @app.get("/redirect")
    def redirect():
        return {"detail": "Redirected"}

    return instance


@pytest.mark.asyncio
async def test_exception_call_cookie_error_user_header_false(clean_manager):
    # setup clean_manager
    clean_manager.use_cookie = True
    clean_manager.use_header = False
    response = Mock(cookies={})
    with pytest.raises(Exception) as exc_info:
        await clean_manager(response)

    assert exc_info.value == InvalidCredentialsException


@pytest.mark.asyncio
async def test_exception_call_raises_no_token_auto_error_off(clean_manager):
    clean_manager.auto_error = False
    # set headers so fastapi internals dont raise an error
    response = Mock(headers={"abc": "abc"})
    with pytest.raises(Exception) as exc_info:
        await clean_manager(response)

    assert exc_info.value == InvalidCredentialsException


@pytest.mark.asyncio
async def test_exception_handling(exception_manager, client, invalid_data):
    invalid_token = exception_manager.create_access_token(data={"sub": invalid_data["username"]})
    resp = await client.get(
        "/private/exception",
        headers={"Authorization": f"Bearer {invalid_token}"}
    )

    assert resp.json()["detail"] == "Redirected"


@pytest.mark.asyncio
async def test_exception_change_no_sub(exception_manager, custom_exception):
    no_sub_token = exception_manager.create_access_token(data={"id": "something"})
    with pytest.raises(custom_exception):
        await exception_manager.get_current_user(no_sub_token)


@pytest.mark.asyncio
async def test_exceptions_change_invalid_token(exception_manager, custom_exception):
    invalid_jwt_token = (
        "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.aGVsb"
        "G8gd29ybGQ.SIr03zM64awWRdPrAM_61QWsZchAtgDV"
        "3pphfHPPWkI"
    )  # this token is taken from pyjwt (https://github.com/jpadilla/pyjwt/blob/master/tests/test_api_jwt.py#L82)
    with pytest.raises(custom_exception):
        await exception_manager.get_current_user(invalid_jwt_token)


@pytest.mark.asyncio
async def test_exceptions_change_user_is_none(exception_manager, custom_exception, invalid_data):
    invalid_user_token = exception_manager.create_access_token(data={"sub": invalid_data["username"]})
    with pytest.raises(custom_exception):
        await exception_manager.get_current_user(invalid_user_token)



import os
from typing import Callable, Awaitable, Dict

from fastapi import FastAPI
from async_asgi_testclient import TestClient
import pytest
from pydantic import BaseModel

from fastapi_login import LoginManager


class User(BaseModel):
    name: str
    surname: str
    email: str
    password: str


@pytest.fixture(scope="module")
def app():
    _app = FastAPI()
    yield _app


@pytest.fixture(scope="session")
def db() -> Dict[str, User]:
    return {
        'john@doe.com': User(
            name="John",
            surname="Doe",
            email="john@doe.com",
            password="hunter2"
        ),
        'sandra@johnson.com': User(
            name='Sandra',
            surname='Johnson',
            email='sandra@johnson.com',
            password='sandra1243'
        )
    }


@pytest.fixture(scope="session")
def load_user_fn(db: dict) -> Callable[[str], User]:
    def load_user(email: str):
        return db.get(email)

    return load_user


@pytest.fixture
def secret() -> str:
    return os.urandom(24).hex()


@pytest.fixture(scope="session")
def token_url() -> str:
    return "/auth/token"


@pytest.fixture(scope='module')
def client(app) -> TestClient:
    return TestClient(app)


@pytest.fixture(scope="session")
def default_data(db) -> dict:
    return {'sub': list(db.keys())[0]}


@pytest.fixture(scope="session")
def invalid_data() -> dict:
    return {"username": "invalid@e.mail", "password": "invalid-pw"}


@pytest.fixture()
def clean_manager(secret, token_url) -> LoginManager:
    """ Return a new LoginManager instance """
    return LoginManager(secret, token_url)


class CustomAuthException(Exception):
    pass


@pytest.fixture
def custom_exception():

    return CustomAuthException



import pytest
from starlette.requests import Request

from fastapi_login import LoginManager
from fastapi_login.exceptions import InvalidCredentialsException


@pytest.fixture(scope="module")
def middleware_manager(app, token_url, load_user_fn):
    instance = LoginManager("secret", "/auth/token")
    instance.user_loader(load_user_fn)
    instance.useRequest(app)

    @app.get("/private/request")
    def private_request_route(request: Request):

        user = request.state.user
        if user is None:
            raise InvalidCredentialsException
        else:
            return {"detail": "Success"}

    return instance


@pytest.mark.asyncio
async def test_middleware_unauthorized(middleware_manager, client):
    resp = await client.get("/private/request")

    assert resp.status_code == 401
    assert resp.json()["detail"] == InvalidCredentialsException.detail


@pytest.mark.asyncio
async def test_middleware_authorized(middleware_manager, client, default_data):
    token = middleware_manager.create_access_token(data=default_data)
    resp = await client.get(
        "/private/request",
        headers={"Authorization": f"Bearer {token}"}
    )

    assert resp.status_code == 200
    assert resp.json()["detail"] == "Success"



import inspect
import typing
from datetime import timedelta, datetime
from typing import Callable, Awaitable, Union

import jwt
from fastapi.security import OAuth2PasswordBearer
from passlib.context import CryptContext
from starlette.datastructures import Secret

from fastapi import FastAPI, Request, Response

from fastapi_login.exceptions import InvalidCredentialsException


class LoginManager(OAuth2PasswordBearer):
    """
    Attributes:
        secret (starlette.datastructures.Secret): The secret used to sign and decrypt the JWT
        algorithm (str): The algorithm used to decrypt the token defaults to ``HS256``
        token_url (str): The url where the user can login to get the token
        use_cookie (bool): Whether cookies should be checked for the token
        use_header (bool): Whether headers should be checked for the token
        pwd_context (passlib.CryptContext): Instance of ``passlib.CryptContext`` using bcrypt for
            convenient access to hashing and verifying passwords.
        cookie_name (str): The name of the cookie checked for the token, defaults to `"access-token"`
    """

    def __init__(self, secret: str, token_url: str, algorithm="HS256", use_cookie=False, use_header=True):
        """
        Initializes LoginManager

        Args:
            algorithm (str): Should be "HS256" or "RS256" used to decrypt the JWT
            token_url (str): The url where the user can login to get the token
            use_cookie (bool): Set if cookies should be checked for the token
            use_header (bool): Set if headers should be checked for the token
        """
        if use_cookie is False and use_header is False:
            raise Exception("use_cookie and use_header are both False one of them needs to be True")
        self.secret = Secret(secret)
        self._user_callback = None
        self.algorithm = algorithm
        self.pwd_context = CryptContext(schemes=["bcrypt"])
        self.tokenUrl = token_url
        self.oauth_scheme = None
        self._not_authenticated_exception = InvalidCredentialsException

        self.use_cookie = use_cookie
        self.use_header = use_header
        self.cookie_name = 'access-token'

        super().__init__(tokenUrl=token_url, auto_error=True)

    @property
    def not_authenticated_exception(self):
        """
        Exception raised when no (valid) token is present.
        Defaults to `fastapi_login.exceptions.InvalidCredentialsException`
        """
        return self._not_authenticated_exception

    @not_authenticated_exception.setter
    def not_authenticated_exception(self, value: Exception):
        """
        Setter for the Exception which raises when the user is not authenticated.
        Sets `self.auto_error` to False in order to raise the correct exception.

        Args:
            value (Exception): The Exception you want to raise
        """
        assert issubclass(value, Exception)  # noqa
        self._not_authenticated_exception = value
        # change auto error setting on OAuth2PasswordBearer
        self.auto_error = False

    def user_loader(self, callback: Union[Callable, Awaitable]) -> Union[Callable, Awaitable]:
        """
        This sets the callback to retrieve the user.
        The function should take an unique identifier like an email
        and return the user object or None.

        Basic usage:

            >>> from fastapi import FastAPI
            >>> from fastapi_login import LoginManager

            >>> app = FastAPI()
            >>> # use import os; print(os.urandom(24).hex()) to get a true secret key
            >>> SECRET = "super-secret"

            >>> manager = LoginManager(SECRET, token_url="Login")

            >>> manager.user_loader(get_user)

            >>> @manager.user_loader
            >>> def get_user():
            ...     # get user logic here

        Args:
            callback (Callable or Awaitable): The callback which returns the user

        Returns:
            The callback
        """
        self._user_callback = callback
        return callback

    async def get_current_user(self, token: str):
        """
        This decodes the jwt based on the secret and on the algorithm
        set on the LoginManager.
        If the token is correctly formatted and the user is found
        the user is returned else this raises `LoginManager.not_authenticated_exception`

        Args:
            token (str): The encoded jwt token

        Returns:
            The user object returned by the instances `_user_callback`

        Raises:
            LoginManager.not_authenticated_exception: The token is invalid or None was returned by `_load_user`
        """
        try:
            payload = jwt.decode(
                token,
                str(self.secret),
                algorithms=[self.algorithm]
            )
            # the identifier should be stored under the sub (subject) key
            user_identifier = payload.get('sub')
            if user_identifier is None:
                raise self.not_authenticated_exception
        # This includes all errors raised by pyjwt
        except jwt.PyJWTError:
            raise self.not_authenticated_exception

        user = await self._load_user(user_identifier)

        if user is None:
            raise self.not_authenticated_exception

        return user

    async def _load_user(self, identifier: typing.Any):
        """
        This loads the user using the user_callback

        Args:
            identifier (Any): The user identifier expected by `_user_callback`

        Returns:
            The user object returned by `_user_callback` or None

        Raises:
            Exception: The user_loader has not been set
        """
        if self._user_callback is None:
            raise Exception(
                "Missing user_loader callback"
            )

        if inspect.iscoroutinefunction(self._user_callback):
            user = await self._user_callback(identifier)
        else:
            user = self._user_callback(identifier)

        return user

    def create_access_token(self, *, data: dict, expires: timedelta = None) -> str:
        """
        Helper function to create the encoded access token using
        the provided secret and the algorithm of the LoginManager instance

        Args:
            data (dict): The data which should be stored in the token
            expires (datetime.timedelta):  An optional timedelta in which the token expires.
                Defaults to 15 minutes

        Returns:
            The encoded JWT with the data and the expiry. The expiry is
            available under the 'exp' key
        """

        to_encode = data.copy()

        if expires:
            expires_in = datetime.utcnow() + expires
        else:
            # default to 15 minutes expiry times
            expires_in = datetime.utcnow() + timedelta(minutes=15)

        to_encode.update({'exp': expires_in})
        encoded_jwt = jwt.encode(to_encode, str(self.secret), self.algorithm)
        # decode here decodes the byte str to a normal str not the token
        return encoded_jwt

    def set_cookie(self, response: Response, token: str) -> None:
        """
        Utility function to set a cookie containing token on the response

        Args:
            response (fastapi.Response): The response which is send back
            token (str): The created JWT
        """
        response.set_cookie(
            key=self.cookie_name,
            value=token,
            httponly=True
        )

    def _token_from_cookie(self, request: Request) -> typing.Optional[str]:
        """
        Checks the requests cookies for cookies with the name `self.cookie_name`

        Args:
            request (fastapi.Request):  The request to the route, normally filled in automatically

        Returns:
            The access token found in the cookies of the request or None

        Raises:
            LoginManager.not_authenticated_exception: No cookie with name ``LoginManager.cookie_name``
                is set on the Request
        """
        token = request.cookies.get(self.cookie_name)

        # we dont use `token is None` in case a cookie with self.cookie_name
        # exists but is set to "", in which case `token is None` evaluates to False
        if not token and self.auto_error:
            # either default InvalidCredentialsException or set by user
            raise self.not_authenticated_exception

        else:
            # Token may be "" so we convert to None
            return token if token else None

    async def __call__(self, request: Request):
        """
        Provides the functionality to act as a Dependency

        Args:
            request (fastapi.Request):The incoming request, this is set automatically
                by FastAPI

        Returns:
            The user object or None

        Raises:
            LoginManager.not_authenticated_exception: If set by the user and `self.auto_error` is set to False

        """
        token = None
        try:
            if self.use_cookie:
                token = self._token_from_cookie(request)
        # The Exception is either a InvalidCredentialsException
        # or a custom exception set by the user
        except Exception as e:
            # In case use_cookie and use_header is enabled
            # headers should be checked if cookie lookup fails
            if self.use_header:
                pass
            else:
                raise e

        if token is None and self.use_header:
            token = await super(LoginManager, self).__call__(request)

        if token is None:
            # No token is present in the request and no Exception has been raised (auto_error=False)
            raise self.not_authenticated_exception
        else:
            return await self.get_current_user(token)

    def useRequest(self, app: FastAPI):
        """
        Add the instance as a middleware, which adds the user object, if present,
        to the request state

        Args:
            app (fastapi.FastAPI): A instance of FastAPI
        """
        @app.middleware("http")
        async def user_middleware(request: Request, call_next):
            try:
                request.state.user = await self.__call__(request)
            except Exception as e:
                # An error occurred while getting the user
                # as middlewares are called for every incoming request
                # it's not a good idea to return the Exception
                # so we set the user to None
                request.state.user = None
            
            return await call_next(request)



import pytest
from fastapi import Depends

from fastapi_login import LoginManager


@pytest.fixture
def header_manager(app, secret, token_url, load_user_fn) -> LoginManager:
    instance = LoginManager(secret, token_url)
    instance.user_loader(load_user_fn)

    @app.get("/private/header")
    def private_header_route(_=Depends(instance)):
        return {"detail": "Success"}

    return instance


@pytest.fixture
def cookie_manager(app, secret, token_url, load_user_fn) -> LoginManager:
    instance = LoginManager(secret, token_url, use_cookie=True, use_header=False)
    instance.user_loader(load_user_fn)

    @app.get("/private/cookie")
    def private_cookie_route(_=Depends(instance)):
        return {"detail": "Success"}

    return instance


@pytest.fixture
def cookie_header_manager(app, secret, token_url, load_user_fn) -> LoginManager:
    instance = LoginManager(secret, token_url, use_cookie=True)
    instance.user_loader(load_user_fn)

    @app.get("/private/both")
    def private_route(_=Depends(instance)):
        return {"detail": "Success"}

    return instance


@pytest.mark.asyncio
async def test_header_dependency(client, header_manager, default_data):
    token = header_manager.create_access_token(data=default_data)
    resp = await client.get(
        "/private/header",
        headers={"Authorization": f"Bearer {token}"}
    )

    assert resp.status_code == 200
    assert resp.json()["detail"] == "Success"


@pytest.mark.asyncio
async def test_cookie_dependency(client, cookie_manager, default_data):
    token = cookie_manager.create_access_token(data=default_data)
    resp = await client.get(
        "/private/cookie",
        cookies={cookie_manager.cookie_name: token}
    )

    assert resp.status_code == 200
    assert resp.json()["detail"] == "Success"


@pytest.mark.asyncio
async def test_cookie_header_fallback(client, cookie_header_manager, default_data):
    token = cookie_header_manager.create_access_token(data=default_data)
    resp = await client.get(
        "/private/both",
        headers={"Authorization": f"Bearer {token}"},
        cookies={}
    )

    # even tough no valid access cookie is present,
    # as use_header is enabled the request is valid
    assert resp.status_code == 200
    assert resp.json()["detail"] == "Success"



import setuptools

with open("README.md", 'r') as f:
    long_description = f.read()


setuptools.setup(
    name="fastapi-login",
    version="1.6.1",
    author="Max Rausch-Dupont",
    author_email="maxrd79@gmail.com",
    descritpion="Flask-Login like package for FastAPI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MushroomMaula/fastapi_login",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    packages=setuptools.find_packages(),
    install_requires=[
        "fastapi",
        "passlib",
        "pyjwt"
    ],
    zip_safe=False,
    include_package_data=True
)